@@ -1,5 +1,4 @@
 /*
-  $Id: extattrs.c,v 1.29 2010-01-05 12:06:33 franklahm Exp $
   Copyright (c) 2009 Frank Lahm <franklahm@gmail.com>
 
   This program is free software; you can redistribute it and/or modify
@@ -29,6 +28,7 @@
 #include <atalk/logger.h>
 #include <atalk/ea.h>
 #include <atalk/globals.h>
+#include <atalk/netatalk_conf.h>
 
 #include "volume.h"
 #include "desktop.h"
@@ -73,7 +73,7 @@ static void hexdump(void *m, size_t l) {
 */
 int afp_listextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf, size_t *rbuflen)
 {
-    int                 ret, oflag = 0, adflags = 0;
+    int                 ret, oflag = 0, adflags = 0, fd = -1;
     uint16_t            vid, bitmap, uint16;
     uint32_t            did, maxreply, tmpattr;
     struct vol          *vol;
@@ -81,11 +81,13 @@ int afp_listextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf,
     struct path         *s_path;
     struct stat         *st;
     struct adouble      ad, *adp = NULL;
+    struct ofork	*opened = NULL;
     char                *uname, *FinderInfo;
     char                emptyFinderInfo[32] = { 0 };
 
     static int          buf_valid = 0;
     static size_t       attrbuflen = 0;
+    bool                close_ad = false;
 
     *rbuflen = 0;
     ibuf += 2;
@@ -139,7 +141,6 @@ int afp_listextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf,
             return( AFPERR_NOOBJ );
         }
 
-        adp = of_ad(vol, s_path, &ad);
         uname = s_path->u_name;
         /*
           We have to check the FinderInfo for the file, because if they aren't all 0
@@ -148,13 +149,24 @@ int afp_listextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf,
           via FPGetExtAttr !
         */
 
-        if (S_ISDIR(st->st_mode))
-            adflags = ADFLAGS_DIR;
+        adp = &ad;
+        ad_init(adp, vol);
 
-        if ( ad_metadata( uname, adflags, adp) < 0 ) {
+        if (path_isadir(s_path)) {
+	    LOG(log_debug, logtype_afpd, "afp_listextattr(%s): is a dir", uname);
+            adflags = ADFLAGS_DIR;
+	} else {
+	    LOG(log_debug, logtype_afpd, "afp_listextattr(%s): is a file", uname);
+	    opened = of_findname(vol, s_path);
+	    if (opened) {
+		adp = opened->of_ad;
+		fd = ad_meta_fileno(adp);
+	    }
+	}
+
+        if (ad_metadata(uname, adflags, adp) != 0 ) {
             switch (errno) {
             case ENOENT:
-                adp = NULL;
                 break;
             case EACCES:
                 LOG(log_error, logtype_afpd, "afp_listextattr(%s): %s: check resource fork permission?",
@@ -164,9 +176,8 @@ int afp_listextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf,
                 LOG(log_error, logtype_afpd, "afp_listextattr(%s): error getting metadata: %s", uname, strerror(errno));
                 return AFPERR_MISC;
             }
-        }
-
-        if (adp) {
+        } else {
+            close_ad = true;
             FinderInfo = ad_entry(adp, ADEID_FINDERI);
             /* Check if FinderInfo equals default and empty FinderInfo*/
             if (memcmp(FinderInfo, emptyFinderInfo, 32) != 0) {
@@ -185,8 +196,8 @@ int afp_listextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf,
                 attrbuflen += strlen(ea_resourcefork) + 1;
             }
         }
-        
-        ret = vol->vfs->vfs_ea_list(vol, attrnamebuf, &attrbuflen, uname, oflag);
+	
+        ret = vol->vfs->vfs_ea_list(vol, attrnamebuf, &attrbuflen, uname, oflag, fd);
 
         switch (ret) {
         case AFPERR_BADTYPE:
@@ -202,7 +213,7 @@ int afp_listextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf,
         default:
             buf_valid = 1;
         }
-    }
+    } /* if ((maxreply == 0) || (buf_valid == 0)) */
 
     /* Start building reply packet */
     bitmap = htons(bitmap);
@@ -228,8 +239,9 @@ int afp_listextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf,
 exit:
     if (ret != AFP_OK)
         buf_valid = 0;
-    if (adp)
-        ad_close_metadata( adp);
+
+    if (close_ad)
+        ad_close(adp, ADFLAGS_HF);
 
     return ret;
 }
@@ -249,13 +261,15 @@ static char attrmname[256];
 
 int afp_getextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf, size_t *rbuflen)
 {
-    int                 ret, oflag = 0;
+    int                 ret, oflag = 0, fd = -1;
     uint16_t            vid, bitmap, attrnamelen;
     uint32_t            did, maxreply;
     char                attruname[256];
     struct vol          *vol;
     struct dir          *dir;
     struct path         *s_path;
+    struct adouble	ad, *adp = NULL;
+    struct ofork	*opened = NULL;
 
 
     *rbuflen = 0;
@@ -316,22 +330,33 @@ int afp_getextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf,
     rbuf += sizeof(bitmap);
     *rbuflen += sizeof(bitmap);
 
+    if (path_isadir(s_path)) {
+	LOG(log_debug, logtype_afpd, "afp_getextattr(%s): is a dir", s_path->u_name);
+    } else {
+	LOG(log_debug, logtype_afpd, "afp_getextattr(%s): is a file", s_path->u_name);
+	opened = of_findname(vol, s_path);
+	if (opened) {
+	    adp = opened->of_ad;
+	    fd = ad_meta_fileno(adp);
+	}
+    }
+
     /*
       Switch on maxreply:
       if its 0 we must return the size of the requested attribute,
       if its non 0 we must return the attribute.
     */
     if (maxreply == 0)
-        ret = vol->vfs->vfs_ea_getsize(vol, rbuf, rbuflen, s_path->u_name, oflag, attruname);
+        ret = vol->vfs->vfs_ea_getsize(vol, rbuf, rbuflen, s_path->u_name, oflag, attruname, fd);
     else
-        ret = vol->vfs->vfs_ea_getcontent(vol, rbuf, rbuflen, s_path->u_name, oflag, attruname, maxreply);
+        ret = vol->vfs->vfs_ea_getcontent(vol, rbuf, rbuflen, s_path->u_name, oflag, attruname, maxreply, fd);
 
     return ret;
 }
 
 int afp_setextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _U_, size_t *rbuflen)
 {
-    int                 oflag = 0, ret;
+    int                 oflag = 0, ret, fd = -1;
     uint16_t            vid, bitmap, attrnamelen;
     uint32_t            did, attrsize;
     char                attruname[256];
@@ -339,6 +364,8 @@ int afp_setextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _
     struct vol          *vol;
     struct dir          *dir;
     struct path         *s_path;
+    struct adouble	ad, *adp = NULL;
+    struct ofork	*opened = NULL;
 
     *rbuflen = 0;
     ibuf += 2;
@@ -378,6 +405,18 @@ int afp_setextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _
         return AFPERR_NOOBJ;
     }
 
+    if (path_isadir(s_path)) {
+	LOG(log_debug, logtype_afpd, "afp_setextattr(%s): is a dir", s_path->u_name);
+    } else {
+	LOG(log_debug, logtype_afpd, "afp_setextattr(%s): is a file", s_path->u_name);
+	opened = of_findname(vol, s_path);
+	if (opened) {
+	    adp = opened->of_ad;
+	    fd = ad_meta_fileno(adp);
+	}
+    }
+
+
     if ((unsigned long)ibuf & 1)
         ibuf++;
 
@@ -404,20 +443,22 @@ int afp_setextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _
 
     LOG(log_debug, logtype_afpd, "afp_setextattr(%s): EA: %s, size: %u", s_path->u_name, to_stringz(attrmname, attrnamelen), attrsize);
 
-    ret = vol->vfs->vfs_ea_set(vol, s_path->u_name, attruname, ibuf, attrsize, oflag);
+    ret = vol->vfs->vfs_ea_set(vol, s_path->u_name, attruname, ibuf, attrsize, oflag, fd);
 
     return ret;
 }
 
 int afp_remextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _U_, size_t *rbuflen)
 {
-    int                 oflag = 0, ret;
+    int                 oflag = 0, ret, fd = -1;
     uint16_t            vid, bitmap, attrnamelen;
     uint32_t            did;
     char                attruname[256];
     struct vol          *vol;
     struct dir          *dir;
     struct path         *s_path;
+    struct adouble	ad, *adp = NULL;
+    struct ofork	*opened = NULL;
 
     *rbuflen = 0;
     ibuf += 2;
@@ -445,10 +486,21 @@ int afp_remextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _
 
     /* get name */
     if (NULL == ( s_path = cname( vol, dir, &ibuf )) ) {
-        LOG(log_debug, logtype_afpd, "afp_setextattr: cname error: %s", strerror(errno));
+        LOG(log_debug, logtype_afpd, "afp_remextattr: cname error: %s", strerror(errno));
         return AFPERR_NOOBJ;
     }
 
+    if (path_isadir(s_path)) {
+	LOG(log_debug, logtype_afpd, "afp_remextattr(%s): is a dir", s_path->u_name);
+    } else {
+	LOG(log_debug, logtype_afpd, "afp_remextattr(%s): is a file", s_path->u_name);
+	opened = of_findname(vol, s_path);
+	if (opened) {
+	    adp = opened->of_ad;
+	    fd = ad_meta_fileno(adp);
+	}
+    }
+
     if ((unsigned long)ibuf & 1)
         ibuf++;
 
@@ -465,7 +517,7 @@ int afp_remextattr(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _
 
     LOG(log_debug, logtype_afpd, "afp_remextattr(%s): EA: %s", s_path->u_name, to_stringz(ibuf, attrnamelen));
 
-    ret = vol->vfs->vfs_ea_remove(vol, s_path->u_name, attruname, oflag);
+    ret = vol->vfs->vfs_ea_remove(vol, s_path->u_name, attruname, oflag, fd);
 
     return ret;
 }
